#if T_TUNE

/* all of source was replaced , based "x1_tape.c" in X1EMU 0.5 */
#include	<windows.h>
#include	<stdio.h>
#include	"x1.h"
#include	"xmil.h"
#include	"x1_cmt.h"
#include	"x1_scpu.h"
#include	"menu.h"
#include	"dosio.h"
#include	"dsounds.h"

/* */
#define tape_freq (tape.header.frequency)
#define tape_pos  (tape.header.position)
#define tape_end  (tape.header.datasize)
#define tape_protect (tape.header.protect)

/* APSS̑x */
#define APSS_SPEED 24/* PLAY24{ */
/* FF/REW̑x */
#define FF_REW_SPEED 24

/* APSŠoԐݒ */
#define APSS_HEAD_IGNORE (tape_freq*2) /* Jn̖ 2sec */
#define APSS_NOSILENT    tape_freq     /* Lo 1sec */
#define APSS_SILENT      tape_freq     /* o 1sec */

CMT_TABLE tape;
static char tape_filename[128] = { 0 };

static EVENT motor_event = EVENT_VALUE(NULL,0,NULL,0);
static EV_TIME motor_interval;

static void cmt_head_up(void);


static const X1TAPE_HEADER default_header =
{
	TAPE_INDEX,
	"auto converted",
	{0,0,0,0,0},
	0,
	TAPE_FORMAT_SAMPLING,
	0,
	8000,
	0
};

/**********************************************************************
	Xe[^X̍쐬
**********************************************************************/
/* update tape state and window title */
char *cmt_get_stat_str(void)
{
	static char work[128];

	unsigned long t;
	int s,m,h;
	static const char *cmt_cmd_str[]=
	{
		"EJT","STP","PLY","FF ",
		"REW","AFF","ARW","   ",
		"   ","   ","REC"
	};

	t=tape_freq ? (tape_pos / tape_freq) : 0;
	s=(int)(t%60l); t/=60l;
	m=(int)(t%60l); t/=60l;
	h=(int)(t%60l);
	sprintf(work," %s(%1d:%02d:%02d)",cmt_cmd_str[tape.cur_cmd],h,m,s);

	return work;
}


void init_cmt(void)
{
	if( tape_filename[0] == 0)
	{
		ZeroMemory(&tape, sizeof(CMT_TABLE));
		tape.cur_cmd = CMT_STOP;
		cmt_set_protect(0);
	}
	else
		cmt_stop();
}

/**********************************************************************
	e[vwb_[̃Z[u
**********************************************************************/
static void cmt_save_header(void)
{
	/* output tape position */
	if( (tape.header.index == TAPE_INDEX) && (tape_freq != 0) )
	{
		FILEH hdr;

		/* write header to file */
		if ((hdr = file_open(tape_filename)) != (FILEH)-1)
		{
			if (file_seek(hdr, 0, FSEEK_SET) == 0)
				file_write(hdr, &tape.header, sizeof(tape.header));
				file_close(hdr);
		}
	}
}

/**********************************************************************
	e[vt@C̓ǂݏ
**********************************************************************/

/* flush curent buffer */
static int cmt_flush_buffer(void)
{
	FILEH hdr;
	DWORD cur_ptr;
	int ret = 0;

	if(!tape_freq)
		return(-1);

	if (!tape.dirty_buf || (tape.page==-1) )
		return(0);

	if ((hdr = file_open(tape_filename)) == (FILEH)-1) {
		return(-1);
	}

	cur_ptr = tape.page * CMT_PAGE_SIZE + sizeof(X1TAPE_HEADER);
	tape.page = -1;
	tape.dirty_buf = 0;
	if ((file_seek(hdr, cur_ptr, FSEEK_SET) != cur_ptr) ||
		(file_write(hdr, tape.buf, tape.buf_size) != tape.buf_size)) {
		ret = -1;
	}
	if (file_close(hdr)) {
		ret = -1;
	}
	return(ret);
}

/* get pointer of tape data */
static BYTE *cmt_data_ptr(DWORD pos)
{
	static BYTE dummy_data=0;
	FILEH hdr;
	DWORD cur_ptr;
	DWORD page;

	if(!tape_freq)
		return(&dummy_data);

	page = pos / CMT_PAGE_SIZE;
	pos %= CMT_PAGE_SIZE;
	if(tape.page == page)
		return &tape.buf[pos];

	/* flush current buffer */
	if( cmt_flush_buffer() < 0)
		return(&dummy_data);

	/* read new buffer */
	if ((hdr = file_open(tape_filename)) == (FILEH)-1)
		return(&dummy_data);

	cur_ptr = page * CMT_PAGE_SIZE + sizeof(X1TAPE_HEADER);

	if(cur_ptr+CMT_PAGE_SIZE > tape.filesize)
		tape.buf_size = (WORD)(tape.filesize - cur_ptr);
	else
		tape.buf_size = CMT_PAGE_SIZE;

	if ((file_seek(hdr, cur_ptr, FSEEK_SET) != cur_ptr) ||
		(file_read(hdr, tape.buf, tape.buf_size) != tape.buf_size))
	{
		file_close(hdr);
		return &dummy_data;
	}

	if (file_close(hdr))
		return &dummy_data;

	tape.page = page;
	return &tape.buf[pos];
}

/**********************************************************************
	ZT[̃`FbN
**********************************************************************/

static void cmt_set_sensor(void)
{
	if(!tape_freq)
		tape.sensor = 0;
	else
	{
		tape.sensor |= 2;    /*  0=ż 1= */
		if(!tape_protect)
			tape.sensor |= 4; /*  1=OK */
		else
			tape.sensor &= ~4; /*  0=ݼ */
	}
}

/**********************************************************************
	e[v[^[̃CxgnhƁA[^[̐
**********************************************************************/

/* set motor event */
static void cmt_motor_start(void (*callback)(int),int param)
{
	/* set motor event interval with adjust overclock rate */
	/* CPŨNbN͎xvZBݒlQƂĂ̓_B */
	motor_event.name     = "CMT MOTOR";
	motor_event.interval =  TIME_IN_CYCLES(4000000L,0) / tape_freq;
	motor_event.callback = callback;
	motor_event.callback_param = param;
	event_set(&motor_event,motor_event.interval);
}

/* adjust timming sampling period to write eddge */
static void adjust_rec_timming(void)
{
	if( tape.cur_cmd == CMT_REC)
	{
		event_set(&motor_event,motor_event.interval/2);
	}
}

/* tape position increment & check break */
static int tape_step_pos(long step)
{
	if(step >= 0)
	{
		if( tape_pos+step >= tape_end)
		{
			/* tape end */
			tape_pos = tape_end-1;
			tape.sensor &= ~1;
			cmt_stop();
			return 1;
		}
		tape_pos += step;
	}
	else
	{
		step = -step;
		if( tape_pos <= (unsigned long)step )
		{
			tape_pos = 0;
			tape.sensor &= ~1;
			cmt_stop();
			return 1;
		}
		tape_pos -= step;
	}
	/* not tape end */
	tape.sensor |= 1;
	return 0;
}

/* PLAY one sampling rate */
static void cmt_play_callback(int param)
{
	tape_step_pos(1);
}

/* REC one sampling rate */
static void cmt_rec_callback(int param)
{
	BYTE bit_mask;
	BYTE *ptr;

	tape_step_pos(1);
	/* write data */
	bit_mask  = 0x80 >> (tape_pos&7);
	ptr = cmt_data_ptr(tape_pos/8);
	if(tape.data)
		(*cmt_data_ptr(tape_pos/8)) |= bit_mask;
	else
		(*cmt_data_ptr(tape_pos/8)) &= ~bit_mask;
	tape.dirty_buf = 1;
}

/* FF/REW sampling rate */
static void cmt_ff_rew_callback(int param)
{
	tape_step_pos( (long)param);
}

/* APSS FF/REW sampling rate */
static void cmt_apss_callback(int speed)
{
	unsigned long pos;
	BYTE data,mask;
	int num_bit =0;

	if( tape_step_pos((long)speed) )
		return;

	/* increment tape position and sampling data bits */
	for( pos=tape_pos ; pos != tape_pos+APSS_SPEED ; pos++)
	{
		mask  = 0x80>>(pos&7);
		data  = *cmt_data_ptr(pos/8);
		if(data&mask)
			num_bit++;
	}

	/* O^1̊APFX܂͂XFP̎AƂ */
	if(num_bit*2 > APSS_SPEED) // 50% ? 
		num_bit = APSS_SPEED-num_bit;
	if( num_bit*5 < APSS_SPEED) // 20%
	{	/* silent */
		if( (tape.apss_cnt>=-(long)APSS_NOSILENT) && (tape.apss_cnt<0) )
			tape.apss_cnt = -(long)APSS_NOSILENT;
	}
	else
	{	/* not silent */
		if(tape.apss_cnt >0)
			tape.apss_cnt = 0;
	}

	/* add silent count & finish check */
	tape.apss_cnt += APSS_SPEED;
	if( tape.apss_cnt >= (long)APSS_SILENT)
		cmt_ctrl(CMT_STOP,0);
}

/* finish HEAD down */
static void cmt_head_down_callback(int param)
{
	int speed;
	int ff_flag = 0;

	tape.busy = 0;
	tape.headdown = 1;

	switch(tape.cur_cmd)
	{
		case CMT_PLAY:
			cmt_motor_start(cmt_play_callback,0);
			/* no wait ! */
			xmilcfg.NOWAIT |= 0x02;
			x1_sub_cmtrx(tape.cur_cmd);
			break;
		case CMT_FF:
		case CMT_REW:
			ff_flag = 1;
			speed = (tape.cur_cmd==CMT_FF) ? FF_REW_SPEED : -FF_REW_SPEED;
			cmt_motor_start(cmt_ff_rew_callback,speed);
			x1_sub_cmtrx(tape.cur_cmd);
			break;
		case CMT_AFF:
		case CMT_AREW:
			ff_flag = 1;
			speed = (tape.cur_cmd==CMT_AFF) ? APSS_SPEED : -APSS_SPEED;
			/* Lo{o̎Ԃ𑫂 */
			tape.apss_cnt = -(long)(APSS_HEAD_IGNORE+APSS_NOSILENT);
			cmt_motor_start(cmt_apss_callback,speed);
			/* no wait ! */
			xmilcfg.NOWAIT |= 0x02;
			x1_sub_cmtrx(tape.cur_cmd);
			break;
		case CMT_REC:
			if(tape_protect) return;
			cmt_motor_start(cmt_rec_callback,0);
			/* no wait ! */
			xmilcfg.NOWAIT |= 0x02;
			x1_sub_cmtrx(tape.cur_cmd);
			break;
		default:
			/* STOP,EJECT */
			cmt_head_up();
	}
	if(ff_flag && xmilcfg.MOTOR)
		WAVE_PLAY(WAV_FFREW,1);
}

/* HEAD down */
static void cmt_head_down(void)
{
	if(!tape_freq)
		return;

	if(!tape.headdown)
	{
		/* ODTb */
		motor_event.name     = "CMT HEAD DOWN";
		motor_event.interval = 0;
		motor_event.callback = cmt_head_down_callback;
		event_set(&motor_event,TIME_IN_MSEC(500));
		tape.busy = 1;

		if(xmilcfg.MOTOR)
			WAVE_PLAY(WAV_PLAY,0);
	}
	else
		cmt_head_down_callback(0);
}

/* finish HEAD up */
static void cmt_head_up_callback(int param)
{
	tape.busy = 0;
	tape.headdown = 0;

	switch(tape.cur_cmd)
	{
		case CMT_EJECT:
			x1_sub_cmtrx(tape.cur_cmd);
			cmt_set_sensor();
			tape.cur_cmd = CMT_STOP;
			if(xmilcfg.MOTOR)
				WAVE_PLAY(WAV_EJECT,0);
			break;
		case CMT_STOP:
			x1_sub_cmtrx(tape.cur_cmd);
			break;
		case CMT_PLAY:
		case CMT_FF:
		case CMT_REW:
		case CMT_AFF:
		case CMT_AREW:
		case CMT_REC:
			if(tape_freq)
				cmt_head_down();
			break;
	}
}

/* HEAD up */
static void cmt_head_up(void)
{
	if(tape.headdown)
	{
		/* ODTb */
		motor_event.name     = "CMT HEAD UP";
		motor_event.interval = 0;
		motor_event.callback = cmt_head_up_callback;
		event_set(&motor_event,TIME_IN_MSEC(500));
		tape.busy = 1;

		if(xmilcfg.MOTOR)
		{
			WAVE_STOP(WAV_FFREW);
			WAVE_PLAY(WAV_STOP,0);
		}
		xmilcfg.NOWAIT &= 0xfd;
	}
	else
	{
		/* łɂĂȀ */
		cmt_head_up_callback(0);
	}
}


/**********************************************************************
	e[v̓^CWFNg {^ ɑ֐
**********************************************************************/

char *cmt_get_tape_name(void)
{
	return tape_filename;
}

// sw ... 0=^ 1=^s
void cmt_set(char *fname)
{
	FILEH hdr;

	if(tape_freq)
		cmt_eject();

	strcpy(tape_filename,fname);
	if ((hdr = file_open(tape_filename)) == (FILEH)-1)
		return;
	/* file length */
	tape.filesize = file_length(hdr);
	if(tape.filesize < sizeof(tape.header) )
		return;

	/* read header */
	file_read(hdr, &tape.header ,sizeof(tape.header));
	file_close(hdr);

	/* when old format file , auto conversion */
	if(tape.header.index != TAPE_INDEX)
	{
		DWORD frq = tape.header.index;
		memcpy(&tape.header,&default_header,sizeof(X1TAPE_HEADER) );
		tape_freq = frq;
		tape_end  = (tape.filesize - sizeof(X1TAPE_HEADER))*8;
#if 1
		/* ϊĂ悢ǂ̊mF */
extern HWND hWndMain;

		if( MessageBox(hWndMain,"\
tH[}bg̃e[vt@CłB\n\
ϊƐ擪RUoCg̃f[^͎܂B\n\
VtH[}bgɎϊ܂H"
			,"mF",MB_YESNO | MB_ICONQUESTION | MB_TASKMODAL ) != IDYES )
		{
			tape.header.index = 0; /* Z[uȂ悤 */
		}
#endif
		/* save the header */
		cmt_save_header();
	}

	/* set tape */
	tape.cur_cmd = CMT_STOP;
	/* set protect bit */
	cmt_set_protect(tape_protect);
	/* sensor change check */
	cmt_set_sensor();

	tape.page=-1;

	/* update display status */
	makecaption(1);
}

void cmt_set_protect(BYTE sw)
{
#if 1
#include	"resource.h"
#define	MFCHECK(a) ((a)?MF_CHECKED:MF_UNCHECKED)
extern HWND hWndMain;
	CheckMenuItem(GetMenu(hWndMain), IDM_CMTPROTECT, MFCHECK(sw));
#endif
	tape_protect = sw ? TAPE_PROTECT : 0;
	/* sensor change check */
	cmt_set_sensor();
}

/**********************************************************************
	e[v̂Prbgo
**********************************************************************/

/* read tape port : bit1 = read data */
BYTE cmt_read(void)
{
	BYTE data,bitmask;

	if(tape.cur_cmd == CMT_PLAY)
	{
		bitmask = 0x80>>(tape_pos&7);

		data =*cmt_data_ptr(tape_pos/8);
		if( data & bitmask)
			return 0x02;
	}
	return 0x00;
}

void cmt_write(BYTE data)
{
	if(tape.data != data)
	{
		adjust_rec_timming();
		tape.data = data;
	}
}

/**********************************************************************
	TubotfbLRg[ׂ̈ɃR[֐
**********************************************************************/

void cmt_ctrl(BYTE cmnd,int isButton)
{
	/* sensor command */
	if( cmnd ==CMT_SENSOR)
	{
		cmt_set_sensor();
		x1_sub_cmtrx(tape.sensor | 0x80);
		return;
	}
	/* status command or , same as current command */
	if( cmnd ==CMT_STATUS || tape.cur_cmd == cmnd)
	{
		x1_sub_cmtrx(tape.cur_cmd);
		return;
	}

	/* if REC mode , update write data */
	cmt_flush_buffer();

	/* check break and send break message */
	if(isButton)
	{
		switch(tape.cur_cmd)
		{
		case CMT_PLAY:
		case CMT_AFF:
		case CMT_AREW:
		case CMT_REC:
			x1_sub_cmtrx(CMT_BREAK);
		}
	}

	/* change current command */
	tape.cur_cmd = cmnd;

	/* if no tape , ignore command */
	if(tape_freq==0)
	{
		x1_sub_cmtrx(CMT_STOP);
		return;
	}

	/* if EJECT , remove tape file */
	if(cmnd == CMT_EJECT)
	{
		cmt_save_header();
		/* clear tape data */
		tape_filename[0] = 0;
		/* */
		tape.header.index = 0;
		tape_freq      = 0;
		tape_pos       = 0;
		tape_end       = 0;
		cmt_set_protect(0);
	}

	/* wbh~ĂAxwbhグĂR}hs */
	cmt_head_up();

	/* update display status */
	makecaption(1);
}

BYTE cmt_busy(void)
{
	return tape.busy;
}

#else //T_TUNE

#include	<windows.h>
#include	"x1.h"

static	BYTE	CMT_CMND = 0;
static	BYTE	CMT_STOP = 1;

void init_cmt(void) {

	CMT_CMND = 0;
	CMT_STOP = 1;
}

BYTE cmt_read(void) { return(0); }

void cmt_write(BYTE dat) { }

void cmt_ctrl(BYTE cmnd)
{
	switch(cmnd){
		case 0x00:				// EJECT

		case 0x01:				// STOP
			CMT_STOP = 1;
			break;
//		case 0x02:				// PLAY
//		case 0x03:				// FF
//		case 0x04:				// REW
//		case 0x05:				// APSS_FF
//		case 0x06:				// APSS_REW
//		case 0x0a:				// REC
//			break;
	}
	CMT_CMND = cmnd;
}

BYTE cmt_test(void) {

	if (CMT_STOP) {
		CMT_STOP = 0;
		return(0);
	}
	return(1);
}
#endif //T_TUNE
